@@ -7,12 +7,19 @@ import logging
 
 class AmFastError(Exception):
     """Base exception for this package."""
-    pass
+    # To get around 2.6 deprecation warning
+    def _get_message(self):
+        return self._message
+    def _set_message(self, message):
+        self._message = message
+    message = property(_get_message, _set_message)
 
 def epoch_from_date(date):
+    """Returns epoch milliseconds."""
     return long(calendar.timegm(date.timetuple()) * 1000)
 
 def date_from_epoch(epoch_secs):
+    """Returns datetime."""
     return datetime.utcfromtimestamp(epoch_secs)
 
 def format_byte_string(byte_string):
