@@ -25,19 +25,19 @@ static int big_endian; // Flag == 1 if architecture is big_endian, == 0 if not
 
 /* Context for encoding. */
 typedef struct {
-    char *buf; // Input buffer
-    int pos; // Current position in input buffer
-    int buf_len; // Length of the buffer
     PyObject *class_def_mapper; // Object for getting a ClassDef for an object.
     PyObject *get_class_def_method_name; // Name of the method to call to retrieve a class def.
     PyObject *apply_attr_vals_method_name; // Name of the method to call to apply attributes
     ObjectContext *string_refs; // Keep track of string references
     ObjectContext *object_refs; // Keep track of object references
     ObjectContext *class_refs; // Keep track of class definitions references
+    char *buf; // Input buffer
+    int pos; // Current position in input buffer
+    int buf_len; // Length of the buffer
 } DecoderContext;
 
-static DecoderContext* _create_decoder_context(void);
-static DecoderContext* _copy_decoder_context(DecoderContext *context);
+static DecoderContext* _create_decoder_context(int amf3);
+static DecoderContext* _copy_decoder_context(DecoderContext *context, int amf3);
 static int _destroy_decoder_context(DecoderContext *context);
 
 // ---- DECODING
@@ -85,8 +85,8 @@ static PyObject* decode_long_string_AMF0(DecoderContext *context);
 static PyObject* decode_date_AMF0(DecoderContext *context);
 static PyObject* decode_xml_AMF0(DecoderContext *context);
 static PyObject* decode_typed_object_AMF0(DecoderContext *context);
-static int decode_headers_AMF0(DecoderContext *context, PyObject *packet);
-static int decode_messages_AMF0(DecoderContext *context, PyObject *packet);
+static PyObject* decode_headers_AMF0(DecoderContext *context);
+static PyObject* decode_messages_AMF0(DecoderContext *context);
 
 // Entry functions
 static PyObject* decode(PyObject *self, PyObject *args, PyObject *kwargs);
@@ -96,7 +96,7 @@ static PyObject* _decode(DecoderContext *context);
 // ------------------------ DECODING CONTEXT ------------------------ //
 
 /* Create a new DecoderContext. */
-static DecoderContext* _create_decoder_context()
+static DecoderContext* _create_decoder_context(int amf3)
 {
     DecoderContext *context;
     context = (DecoderContext*)malloc(sizeof(DecoderContext));
@@ -105,32 +105,52 @@ static DecoderContext* _create_decoder_context()
         return NULL;
     }
 
+    // Set python pointers to NULL
     context->buf = NULL;
-    context->buf_len = 0;
-    context->pos = 0;
-
-    context->string_refs = create_object_context(64);
-    if (!context->string_refs) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
-    context->object_refs = create_object_context(64);
-    if (!context->object_refs) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
-    context->class_refs = create_object_context(64);
-    if (!context->class_refs) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
     context->class_def_mapper = NULL;
     context->get_class_def_method_name = NULL;
     context->apply_attr_vals_method_name = NULL;
 
+    if (amf3) {
+        context->string_refs = create_object_context(64);
+        if (!context->string_refs) {
+            free(context);
+            return NULL;
+        }
+    } else {
+        context->string_refs = NULL;
+    }
+
+    // Always create object refs.
+    context->object_refs = create_object_context(64);
+    if (!context->object_refs) {
+        if (context->string_refs) {
+            destroy_object_context(context->string_refs);
+        }
+        free(context);
+        return NULL;
+    }
+
+    if (amf3) {
+        context->class_refs = create_object_context(64);
+        if (!context->class_refs) {
+            if (context->string_refs) {
+                destroy_object_context(context->string_refs);
+            }
+            if (context->object_refs) {
+                destroy_object_context(context->object_refs);
+            }
+
+            free(context);
+            return NULL;
+        }
+    } else {
+        context->class_refs = NULL;
+    }
+
+    context->buf_len = 0;
+    context->pos = 0;
+
     return context;
 }
 
@@ -139,9 +159,9 @@ static DecoderContext* _create_decoder_context()
  *
  * Use this when you need to reset reference counts.
  */
-static DecoderContext* _copy_decoder_context(DecoderContext *context)
+static DecoderContext* _copy_decoder_context(DecoderContext *context, int amf3)
 {
-    DecoderContext *new_context = _create_decoder_context();
+    DecoderContext *new_context = _create_decoder_context(amf3);
     if (!new_context)
         return NULL;
 
@@ -167,9 +187,17 @@ static DecoderContext* _copy_decoder_context(DecoderContext *context)
 /* De-allocate an DecoderContext. */
 static int _destroy_decoder_context(DecoderContext *context)
 {
-    destroy_object_context(context->string_refs);
-    destroy_object_context(context->object_refs);
-    destroy_object_context(context->class_refs);
+    if (context->string_refs) {
+        destroy_object_context(context->string_refs);
+    }
+
+    if (context->object_refs) {
+        destroy_object_context(context->object_refs);
+    }
+
+    if (context->class_refs) {
+        destroy_object_context(context->class_refs);
+    }
 
     if (context->class_def_mapper) {
         Py_DECREF(context->class_def_mapper);
@@ -311,7 +339,7 @@ static int decode_typed_object(DecoderContext *context, PyObject *obj_value, PyO
         return 0;
     }
 
-    int static_attr_len = PyTuple_GET_SIZE(static_attrs);
+    Py_ssize_t static_attr_len = PyTuple_GET_SIZE(static_attrs);
     int i;
     for (i = 0; i < static_attr_len; i++) {
         PyObject *obj = _decode(context);
@@ -321,7 +349,7 @@ static int decode_typed_object(DecoderContext *context, PyObject *obj_value, PyO
             return 0;
         }
 
-        PyObject *attr_name = PyTuple_GetItem(static_attrs, i);
+        PyObject *attr_name = PyTuple_GET_ITEM(static_attrs, i);
         if (!attr_name) {
             Py_DECREF(decoded_attrs);
             Py_DECREF(static_attrs);
@@ -383,9 +411,9 @@ static int decode_externizeable(DecoderContext *context, PyObject *obj_value, Py
 
     #ifdef Py_BYTEARRAYOBJECT_H
     // ByteArray decoding is only available in 2.6+
-    byte_array = PyByteArray_FromStringAndSize(char_value, byte_len);
+    byte_array = PyByteArray_FromStringAndSize(char_value, (Py_ssize_t)byte_len);
     #else
-    byte_array = PyString_FromStringAndSize(char_value, byte_len);
+    byte_array = PyString_FromStringAndSize(char_value, (Py_ssize_t)byte_len);
     #endif
 
     if (!byte_array) {
@@ -400,7 +428,7 @@ static int decode_externizeable(DecoderContext *context, PyObject *obj_value, Py
     if (!parsed_len)
         return 0;
 
-    context->pos = PyInt_AsLong(parsed_len) + 1;
+    context->pos = (int)(PyInt_AsLong(parsed_len) + 1);
     Py_DECREF(parsed_len);
     return 1; 
 }
@@ -444,21 +472,30 @@ static PyObject* decode_class_def(DecoderContext *context, int header)
     if (!alias)
         return NULL;
 
+    // Get alias as C-String for error handling
+    PyObject *alias_str = PyUnicode_AsASCIIString(alias);
+    if (!alias_str) {
+        Py_DECREF(alias);
+        return NULL;
+    }
+    char *alias_char = PyString_AsString(alias_str);
+
     PyObject *class_def = class_def_from_alias(context, alias);
     Py_DECREF(alias);
     if (!class_def) 
         return NULL;
 
     // Check for an externizeable class def.
+    if(PyObject_HasAttrString(class_def, "EXTERNIZEABLE_CLASS_DEF")) {
+        return(class_def);
+    }
+ 
     if ((header & 0x07FFFFFF) == EXTERNIZEABLE) {
-        if(PyObject_HasAttrString(class_def, "EXTERNIZEABLE_CLASS_DEF")) {
-            // Externizeable
-            return(class_def);
-        } else {
-            Py_DECREF(class_def);
-            PyErr_SetString(amfast_DecodeError, "Encoded class is externizeable, but ClassDef is not.");
-            return NULL;
-        }
+        Py_DECREF(class_def);
+        char error_str[512];
+        sprintf(error_str, "Encoded class '%s' is externizeable, but ClassDef is not.", alias_char);
+        PyErr_SetString(amfast_DecodeError, error_str);
+        return NULL;
     }
 
     // Check for anonymous object
@@ -469,11 +506,15 @@ static PyObject* decode_class_def(DecoderContext *context, int header)
     // but encoding is static
     if (PyObject_HasAttrString(class_def, "DYNAMIC_CLASS_DEF") && (!((header & DYNAMIC) == DYNAMIC))) {
         Py_DECREF(class_def);
-        PyErr_SetString(amfast_DecodeError, "Encoded class is static, but ClassDef is dynamic.");
+        char error_str[512];
+        sprintf(error_str, "Encoded class '%s' is static, but ClassDef is dynamic.", alias_char);
+        PyErr_SetString(amfast_DecodeError, error_str);
         return NULL;
     } else if ((header & DYNAMIC) == DYNAMIC && (!PyObject_HasAttrString(class_def, "DYNAMIC_CLASS_DEF"))) {
         Py_DECREF(class_def);
-        PyErr_SetString(amfast_DecodeError, "Encoded class is dynamic, but ClassDef is static.");
+        char error_str[512];
+        sprintf(error_str, "Encoded class '%s' is dynamic, but ClassDef is static.", alias_char);
+        PyErr_SetString(amfast_DecodeError, error_str);
         return NULL;
     }
 
@@ -573,6 +614,9 @@ static int _decode_dynamic_dict(DecoderContext *context, PyObject *dict)
             Py_DECREF(val);
             return 0;
         }
+
+        Py_DECREF(key);
+        Py_DECREF(val);
     }
 }
 
@@ -697,6 +741,13 @@ static PyObject* decode_date(DecoderContext *context)
     if (!epoch_float)
         return NULL;
 
+    if (!amfast_mod) {
+        amfast_mod = PyImport_ImportModule("amfast");
+        if(!amfast_mod) {
+            return NULL;
+        }
+    }
+
     PyObject *func = PyObject_GetAttrString(amfast_mod, "date_from_epoch");
     if (!func) {
         Py_DECREF(epoch_float);
@@ -748,9 +799,9 @@ static PyObject* decode_byte_array(DecoderContext *context, int byte_len)
 
     #ifdef Py_BYTEARRAYOBJECT_H
     // ByteArray decoding is only available in 2.6+
-    byte_array_value = PyByteArray_FromStringAndSize(char_value, byte_len);
+    byte_array_value = PyByteArray_FromStringAndSize(char_value, (Py_ssize_t)byte_len);
     #else
-    byte_array_value = PyString_FromStringAndSize(char_value, byte_len);
+    byte_array_value = PyString_FromStringAndSize(char_value, (Py_ssize_t)byte_len);
     #endif
 
     if (!byte_array_value)
@@ -822,7 +873,10 @@ static PyObject* deserialize_string(DecoderContext *context)
 
     // Check for null string
     if (header == EMPTY_STRING_TYPE) {
-        return PyString_FromStringAndSize(NULL, 0);
+        PyObject *empty_string = PyString_FromStringAndSize(NULL, 0);
+        PyObject *return_value = PyUnicode_FromObject(empty_string);
+        Py_DECREF(empty_string);
+        return return_value;
     }
 
     // Check for reference
@@ -893,20 +947,19 @@ static double _decode_double(DecoderContext *context)
         double d_value;
         char c_value[8];
     } d;
-    char *char_value = d.c_value;
 
     if (big_endian) {
-        memcpy(char_value, context->buf + context->pos, 8);
+        memcpy(d.c_value, context->buf + context->pos, 8);
     } else {
         // Flip endianness
-        char_value[0] = context->buf[context->pos + 7];
-        char_value[1] = context->buf[context->pos + 6],
-        char_value[2] = context->buf[context->pos + 5],
-        char_value[3] = context->buf[context->pos + 4],
-        char_value[4] = context->buf[context->pos + 3],
-        char_value[5] = context->buf[context->pos + 2],
-        char_value[6] = context->buf[context->pos + 1],
-        char_value[7] = context->buf[context->pos];
+        d.c_value[0] = context->buf[context->pos + 7];
+        d.c_value[1] = context->buf[context->pos + 6],
+        d.c_value[2] = context->buf[context->pos + 5],
+        d.c_value[3] = context->buf[context->pos + 4],
+        d.c_value[4] = context->buf[context->pos + 3],
+        d.c_value[5] = context->buf[context->pos + 2],
+        d.c_value[6] = context->buf[context->pos + 1],
+        d.c_value[7] = context->buf[context->pos];
     }
     context->pos += 8;
 
@@ -921,14 +974,13 @@ static unsigned short _decode_ushort(DecoderContext *context)
         unsigned short s_value;
         char c_value[2];
     } s;
-    char *char_value = s.c_value;
 
     if (big_endian) {
-        memcpy(char_value, context->buf + context->pos, 2);
+        memcpy(s.c_value, context->buf + context->pos, 2);
     } else {
         // Flip endianness
-        char_value[0] = context->buf[context->pos + 1];
-        char_value[1] = context->buf[context->pos];
+        s.c_value[0] = context->buf[context->pos + 1];
+        s.c_value[1] = context->buf[context->pos];
     }
     context->pos += 2;
 
@@ -943,23 +995,21 @@ static unsigned int _decode_ulong(DecoderContext *context)
         unsigned int i_value;
         char c_value[4];
     } i;
-    char *char_value = i.c_value;
 
     if (big_endian) {
-        memcpy(char_value, context->buf + context->pos, 4);
+        memcpy(i.c_value, context->buf + context->pos, 4);
     } else {
         // Flip endianness
-        char_value[0] = context->buf[context->pos + 3];
-        char_value[1] = context->buf[context->pos + 2];
-        char_value[2] = context->buf[context->pos + 1];
-        char_value[3] = context->buf[context->pos];
+        i.c_value[0] = context->buf[context->pos + 3];
+        i.c_value[1] = context->buf[context->pos + 2];
+        i.c_value[2] = context->buf[context->pos + 1];
+        i.c_value[3] = context->buf[context->pos];
     }
     context->pos += 4;
 
     return i.i_value;
 }
 
-
 /* Decode a double to a PyFloat. */
 static PyObject* decode_double(DecoderContext *context)
 {
@@ -1001,7 +1051,7 @@ static int _decode_int(DecoderContext *context)
 /* Decode an int to a PyInt. */
 static PyObject* decode_int(DecoderContext *context)
 {
-    return PyInt_FromLong(_decode_int(context)); 
+    return PyInt_FromLong((long)_decode_int(context)); 
 }
 
 /* Decode an AMF0 Boolean. */
@@ -1037,7 +1087,7 @@ static PyObject* decode_long_string_AMF0(DecoderContext *context)
 static PyObject* decode_reference_AMF0(DecoderContext *context)
 {
     unsigned short idx = _decode_ushort(context);
-    return get_ref_from_idx(context->object_refs, idx);
+    return get_ref_from_idx(context->object_refs, (int)idx);
 }
 
 /* Decode an AMF0 dict. */
@@ -1210,11 +1260,9 @@ static PyObject* decode_typed_object_AMF0(DecoderContext *context)
     return obj_value;
 }
 
-/* Decode a AMF0 NetConnection packet. */
+/* Decode an AMF0 NetConnection packet. */
 static PyObject* _decode_packet(DecoderContext *context)
 {
-    PyObject *client_type;
-
     if (!remoting_mod) {
         remoting_mod = PyImport_ImportModule("amfast.remoting");
         if(!remoting_mod)
@@ -1225,109 +1273,108 @@ static PyObject* _decode_packet(DecoderContext *context)
     if (!packet_class)
         return NULL;
 
-    PyObject *packet = PyObject_CallFunctionObjArgs(packet_class, NULL);
-    Py_DECREF(packet_class);
-    if (!packet)
-        return NULL;
-
     // Set client type
+    PyObject *client_type;
     unsigned short amf_version = _decode_ushort(context);
     if (amf_version == FLASH_8) {
-        client_type = PyObject_GetAttrString(packet, "FLASH_8"); 
+        client_type = PyObject_GetAttrString(packet_class, "FLASH_8"); 
     } else if (amf_version == FLASH_COM) {
-        client_type = PyObject_GetAttrString(packet, "FLASH_COM");
+        client_type = PyObject_GetAttrString(packet_class, "FLASH_COM");
     } else if (amf_version == FLASH_9) {
-        client_type = PyObject_GetAttrString(packet, "FLASH_9");
+        client_type = PyObject_GetAttrString(packet_class, "FLASH_9");
     } else {
         PyErr_SetString(amfast_DecodeError, "Unknown client type.");
-        Py_DECREF(packet);
+        Py_DECREF(packet_class);
         return NULL;
     }
 
     if (!client_type) {
-        Py_DECREF(packet);
+        Py_DECREF(packet_class);
         return NULL;
     }
 
-    if (PyObject_SetAttrString(packet, "version", client_type) == -1) {
-        Py_DECREF(packet);
+    PyObject *headers = decode_headers_AMF0(context);
+    if (!headers) {
+        Py_DECREF(packet_class);
+        Py_DECREF(client_type);
         return NULL;
     }
 
-    // Parse headers
-    if (!decode_headers_AMF0(context, packet)) {
-        Py_DECREF(packet);
-        return NULL;
-    }
-
-    // Parse messages
-    if (!decode_messages_AMF0(context, packet)) {
-        Py_DECREF(packet);
+    PyObject *messages = decode_messages_AMF0(context);
+    if (!messages) {
+        Py_DECREF(packet_class);
+        Py_DECREF(client_type);
+        Py_DECREF(headers);
         return NULL;
     }
 
+    PyObject *packet = PyObject_CallFunctionObjArgs(packet_class, client_type, headers, messages, NULL);
+    Py_DECREF(packet_class);
+    Py_DECREF(client_type);
+    Py_DECREF(headers);
+    Py_DECREF(messages);
     return packet;
 }
 
 /* Decode AMF0 packet headers. */
-static int decode_headers_AMF0(DecoderContext *context, PyObject *packet)
+static PyObject* decode_headers_AMF0(DecoderContext *context)
 {
     unsigned short header_count = _decode_ushort(context);
+    PyObject *header_list = PyList_New(header_count);
+    if (!header_list)
+        return NULL;
+
     int i;
     for (i = 0; i < header_count; i++) {
-
         PyObject *header_name = decode_string_AMF0(context);
-        if (!header_name)
-            return 0;
+        if (!header_name) {
+            Py_DECREF(header_list);
+            return NULL;
+        }
 
         PyObject *required = decode_bool_AMF0(context);
         if (!required) {
+            Py_DECREF(header_list);
             Py_DECREF(header_name);
-            return 0;
+            return NULL;
         }
 
         int byte_len = _decode_ulong(context); // Byte length of header.
 
         // We need a new context for each header
-        DecoderContext *new_context = _copy_decoder_context(context);
+        DecoderContext *new_context = _copy_decoder_context(context, 0);
         if (!new_context) {
+            Py_DECREF(header_list);
             Py_DECREF(header_name);
             Py_DECREF(required);
-            return 0;
-        }
-
-        PyObject *header_obj;
-        if (new_context->buf[new_context->pos] == AMF3_AMF0) {
-            // We could rely on _decode_AMF0
-            // to notice the AMF3 byte, but then
-            // we create/destroy the new context
-            // twice.
-            new_context->pos++;
-            header_obj = _decode(new_context);
-        } else {
-            header_obj = _decode_AMF0(new_context);
+            return NULL;
         }
 
+        PyObject *header_obj = _decode_AMF0(new_context);
         context->pos = new_context->pos;
 
         if (!_destroy_decoder_context(new_context)) {
+            Py_DECREF(header_list);
             Py_DECREF(header_name);
             Py_DECREF(required);
-            return 0;
+            Py_XDECREF(header_obj);
+            return NULL;
         }
 
         if (!header_obj) {
+            Py_DECREF(header_list);
             Py_DECREF(header_name);
             Py_DECREF(required);
-            return 0;
+            return NULL;
         }
 
         // Create header object
         PyObject *header_class = PyObject_GetAttrString(remoting_mod, "Header");
         if (!header_class) {
+            Py_DECREF(header_list);
             Py_DECREF(header_name);
             Py_DECREF(required);
-            return 0;
+            return NULL;
         }
 
         PyObject *header = PyObject_CallFunctionObjArgs(header_class, header_name, required, header_obj, NULL);
@@ -1336,88 +1383,77 @@ static int decode_headers_AMF0(DecoderContext *context, PyObject *packet)
         Py_DECREF(required);
         Py_DECREF(header_obj);
         if (!header) {
-            Py_DECREF(header_name);
-            return 0;
+            Py_DECREF(header_list);
+            return NULL;
         }
 
-        // Set header into packet
-        PyObject *header_list = PyObject_GetAttrString(packet, "headers");
-        if (!header_list) {
-            Py_DECREF(header_name);
-            Py_DECREF(header);
-            return 0;
-        }
-
-        int return_value = PyList_Append(header_list, header);
-        Py_DECREF(header);
-        Py_DECREF(header_list);
-        if (return_value == -1) {
-            return 0;
-        }
+        // Steals reference to header
+        PyList_SET_ITEM(header_list, i, header);
     }
 
-    return 1;
+    return header_list;
 }
 
 /* Decode AMF0 packet messages. */
-static int decode_messages_AMF0(DecoderContext *context, PyObject *packet)
+static PyObject* decode_messages_AMF0(DecoderContext *context)
 {
     unsigned short message_count = _decode_ushort(context);
+    PyObject *message_list = PyList_New(message_count);
+    if (!message_list)
+        return NULL;
+
     int i;
     for (i = 0; i < message_count; i++) {
         PyObject *target = decode_string_AMF0(context);
-        if (!target)
-            return 0;
+        if (!target) {
+            Py_DECREF(message_list);
+            return NULL;
+        }
 
         PyObject *response = decode_string_AMF0(context);
         if (!response) {
+            Py_DECREF(message_list);
             Py_DECREF(target);
-            return 0;
+            return NULL;
         }
 
         int byte_len = _decode_ulong(context); // Message byte length
 
         // We need a new context for each message
-        DecoderContext *new_context = _copy_decoder_context(context);
+        DecoderContext *new_context = _copy_decoder_context(context, 0);
         if (!new_context) {
+            Py_DECREF(message_list);
             Py_DECREF(target);
             Py_DECREF(response);
-            return 0;
-        }
-
-        PyObject *message_obj;
-        if (new_context->buf[new_context->pos + 1] == AMF3_AMF0) {
-            // We could rely on _decode_AMF0
-            // to notice the AMF3 byte, but then
-            // we create/destroy the new context
-            // twice.
-            new_context->pos++;
-            message_obj = _decode(new_context);
-        } else {
-            message_obj = _decode_AMF0(new_context);
+            return NULL;
         }
 
+        PyObject *message_obj = message_obj = _decode_AMF0(new_context);
         context->pos = new_context->pos;
 
         if (!_destroy_decoder_context(new_context)) {
+            Py_DECREF(message_list);
             Py_DECREF(target);
             Py_DECREF(response);
-            return 0;
+            Py_XDECREF(message_obj);
+            return NULL;
         }
 
         if (!message_obj) {
+            Py_DECREF(message_list);
             Py_DECREF(target);
             Py_DECREF(response);
-            return 0;
+            return NULL;
         }
 
         // Create message object
         PyObject *message_class = PyObject_GetAttrString(remoting_mod, "Message");
         if (!message_class) {
+            Py_DECREF(message_list);
             Py_DECREF(target);
             Py_DECREF(response);
             Py_DECREF(message_obj);
-            return 0;
+            return NULL;
         }
 
         PyObject *message = PyObject_CallFunctionObjArgs(message_class, target, response, message_obj, NULL);
@@ -1425,35 +1461,31 @@ static int decode_messages_AMF0(DecoderContext *context, PyObject *packet)
         Py_DECREF(target);
         Py_DECREF(response);
         Py_DECREF(message_obj);
-        if (!message)
-            return 0;
-
-        // Set message into packet
-        PyObject *message_list = PyObject_GetAttrString(packet, "messages");
-        if (!message_list) {
-            Py_DECREF(message);
-            return 0;
+        if (!message) {
+            Py_DECREF(message_list);
+            return NULL;
         }
 
-        int return_value = PyList_Append(message_list, message);
-        Py_DECREF(message_list);
-        Py_DECREF(message);
-        if (return_value == -1) {
-            return 0;
-        }
+        // Steals reference
+        PyList_SET_ITEM(message_list, i, message);
     }
 
-    return 1;
+    return message_list;
 }
 
 /* Decode individual AMF0 objects from buffer. */
 static PyObject* _decode_AMF0(DecoderContext *context)
 {
+    if (context->pos > context->buf_len) {
+        PyErr_SetString(amfast_DecodeError, "Parsed past end of buffer.");
+        return NULL;
+    }
+
     char byte = context->buf[context->pos];
 
     if (byte == AMF3_AMF0) {
         // AMF3 item requires new reference counts.
-        DecoderContext *new_context = _copy_decoder_context(context);
+        DecoderContext *new_context = _copy_decoder_context(context, 1);
         new_context->pos++;
         PyObject *return_value = _decode(new_context);
         context->pos = new_context->pos;
@@ -1501,15 +1533,20 @@ static PyObject* _decode_AMF0(DecoderContext *context)
         return decode_typed_object_AMF0(context);
     }
 
-    char error_str[40];
+    char error_str[100];
     sprintf(error_str, "Unknown AMF0 type marker byte: '%X' at position: %d", byte, context->pos);
     PyErr_SetString(amfast_DecodeError, error_str);
     return NULL;
 }
 
-/* Decode individual objects from buffer. */
+/* Decode individual AMF3 objects from buffer. */
 static PyObject* _decode(DecoderContext *context)
 {
+    if (context->pos > context->buf_len) {
+        PyErr_SetString(amfast_DecodeError, "Parsed past end of buffer.");
+        return NULL;
+    }
+
     char byte = context->buf[context->pos];
 
     if (byte == NULL_TYPE) {
@@ -1547,7 +1584,7 @@ static PyObject* _decode(DecoderContext *context)
         return deserialize_date(context);
     }
 
-    char error_str[40];
+    char error_str[100];
     sprintf(error_str, "Unknown AMF3 type marker byte: '%X' at position: %d", byte, context->pos);
     PyErr_SetString(amfast_DecodeError, error_str);
     return NULL;
@@ -1567,15 +1604,19 @@ static PyObject* decode(PyObject *self, PyObject *args, PyObject *kwargs)
         &value, &packet, &amf3, &class_def_mapper))
         return NULL;
 
-    DecoderContext *context = _create_decoder_context();
-    if (!context)
+    DecoderContext *context = _create_decoder_context(amf3);
+    if (!context) {
+        Py_DECREF(class_def_mapper);
         return NULL;
+    }
 
+    context->buf_len = (int)PyString_GET_SIZE(value);
     context->buf = PyString_AsString(value);
-    if (!context->buf)
+    if (!context->buf) {
+        Py_DECREF(class_def_mapper);
+        _destroy_decoder_context(context);
         return NULL;
-
-    context->buf_len = PyString_GET_SIZE(value);
+    }
 
     // Set defaults
     if (class_def_mapper != Py_None) {
@@ -1585,21 +1626,27 @@ static PyObject* decode(PyObject *self, PyObject *args, PyObject *kwargs)
         Py_DECREF(Py_None);
     } else {
         // Create anonymous ClassDefMapper
+        Py_DECREF(Py_None);
         if (!class_def_mod) {
             class_def_mod = PyImport_ImportModule("amfast.class_def");
-            if(!class_def_mod)
+            if(!class_def_mod) {
+                _destroy_decoder_context(context);
                 return NULL;
+            }
         }
 
         PyObject *class_def = PyObject_GetAttrString(class_def_mod, "ClassDefMapper");
-        if (!class_def)
+        if (!class_def) {
+            _destroy_decoder_context(context);
             return NULL;
+        }
 
         context->class_def_mapper = PyObject_CallFunctionObjArgs(class_def, NULL);
         Py_DECREF(class_def);
-        if (!context->class_def_mapper)
+        if (!context->class_def_mapper) {
+            _destroy_decoder_context(context);
             return NULL;
-        Py_DECREF(Py_None);
+        }
     }
 
     PyObject *return_value;
@@ -1620,7 +1667,7 @@ static PyMethodDef decoder_methods[] = {
     {"decode", (PyCFunction)decode, METH_VARARGS | METH_KEYWORDS,
     "Description:\n"
     "=============\n"
-    "Decode an AMF3 stream to Python objects.\n\n"
+    "Decode an AMF stream to Python objects.\n\n"
     "Useage:\n"
     "===========\n"
     "py_obj = decode(value, **kwargs)\n\n"
@@ -1628,7 +1675,7 @@ static PyMethodDef decoder_methods[] = {
     "============================\n"
     " * packet - bool - True to decode as an AMF NetConnection packet (decode a remoting call). - Default = False\n"
     " * amf3 - bool - True to decode as AMF3 format. - Default = False\n"
-    " * class_def_mapper - ClassDefMapper - object that retrieves a ClassDef object used for customizing \n"
+    " * class_def_mapper - ClassDefMapper - object that retrieves ClassDef objects used for customizing \n"
     "    de-serialization of objects - Default = None (all objects are anonymous)\n"},
     {NULL, NULL, 0, NULL}   /* sentinel */
 };
@@ -1665,14 +1712,6 @@ initdecoder(void)
         return;
     }
 
-    // Include class def module
-    if (!class_def_mod) {
-        class_def_mod = PyImport_ImportModule("amfast.class_def");
-        if(!class_def_mod) {
-            return;
-        }
-    }
-
     // Determine endianness of architecture
     if (is_bigendian()) {
         big_endian = 1;
