@@ -26,13 +26,6 @@ static int big_endian; // Flag == 1 if architecture is big_endian, == 0 if not
 
 /* Context for encoding. */
 typedef struct {
-    char *buf; // Output buffer
-    size_t buf_len; // Current length of string in output buffer
-    size_t buf_size; // Current size of output buffer
-    int use_array_collections; // Flag == 1 to encode lists and tuples to ArrayCollection
-    int use_object_proxies; // Flag == 1 to encode dicts to ObjectProxy
-    int use_references; // Flag == 1 to encode multiply occuring objects as references
-    int use_legacy_xml; // Flag == 1 to encode XML as legacy XMLDocument instead of E4X
     PyObject *include_private; // PyBool, if True encode attributes starting with '_'.
     PyObject *class_def_mapper; // Object for getting a ClassDef for an object.
     PyObject *get_class_def_method_name; // Name of the method to call to retrieve a class def.
@@ -41,10 +34,18 @@ typedef struct {
     ObjectContext *string_refs; // Keep track of string references
     ObjectContext *object_refs; // Keep track of object references
     ObjectContext *class_refs; // Keep track of class definitions references
+    char *buf; // Output buffer
+    size_t buf_len; // Current length of string in output buffer
+    size_t buf_size; // Current size of output buffer
+    int use_array_collections; // Flag == 1 to encode lists and tuples to ArrayCollection
+    int use_object_proxies; // Flag == 1 to encode dicts to ObjectProxy
+    int use_references; // Flag == 1 to encode multiply occuring objects as references
+    int use_legacy_xml; // Flag == 1 to encode XML as legacy XMLDocument instead of E4X
+ 
 } EncoderContext;
 
-static EncoderContext* _create_encoder_context(size_t size);
-static EncoderContext* _copy_encoder_context(EncoderContext *context);
+static EncoderContext* _create_encoder_context(size_t size, int amf3);
+static EncoderContext* _copy_encoder_context(EncoderContext *context, int amf3);
 static int _destroy_encoder_context(EncoderContext *context);
 static int _increase_buffer_size(EncoderContext *context, size_t size);
 static int _amf_write_string_size(EncoderContext *context, char *value, size_t size);
@@ -73,7 +74,6 @@ static int encode_unicode(EncoderContext *context, PyObject *value);
 static int serialize_string(EncoderContext *context, PyObject *value);
 static int encode_string(EncoderContext *context, PyObject *value);
 static int serialize_string_or_unicode(EncoderContext *context, PyObject *value);
-static int _serialize_string(EncoderContext *context, char *value);
 static int write_tuple(EncoderContext *context, PyObject *value);
 static int serialize_tuple(EncoderContext *context, PyObject *value);
 static int encode_tuple(EncoderContext *context, PyObject *value);
@@ -137,7 +137,7 @@ static int _encode_object(EncoderContext *context, PyObject *value);
 // ------------------------ ENCODING CONTEXT ------------------------ //
 
 /* Create a new EncoderContext. */
-static EncoderContext* _create_encoder_context(size_t size)
+static EncoderContext* _create_encoder_context(size_t size, int amf3)
 {
     EncoderContext *context;
     context = (EncoderContext*)malloc(sizeof(EncoderContext));
@@ -146,40 +146,62 @@ static EncoderContext* _create_encoder_context(size_t size)
         return NULL;
     }
 
-    context->buf_size = size; // initial buffer size.
-    context->buf_len = 0; // Nothing is in the buffer yet
-    context->buf = (char*)malloc(sizeof(char*) * context->buf_size); // Create buffer
-    if (!context->buf) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
-    context->string_refs = create_object_context(64);
-    if (!context->string_refs) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
-    context->object_refs = create_object_context(64);
-    if (!context->object_refs) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
-    context->class_refs = create_object_context(64);
-    if (!context->class_refs) {
-        PyErr_SetNone(PyExc_MemoryError);
-        return NULL;
-    }
-
+    // Set python object pointers to NULL
     context->include_private = NULL;
     context->class_def_mapper = NULL;
     context->get_class_def_method_name = NULL;
     context->array_collection_def = NULL;
     context->object_proxy_def = NULL;
 
-    context->use_references = 1;
+    // Setup buffer
+    context->buf_size = size; // initial buffer size.
+    context->buf_len = 0; // Nothing is in the buffer yet
+    context->buf = (char*)malloc(sizeof(char*) * context->buf_size); // Create buffer
+    if (!context->buf) {
+        PyErr_SetNone(PyExc_MemoryError);
+        free(context);
+        return NULL;
+    }
 
+    if (amf3) {
+        context->string_refs = create_object_context(64);
+        if (!context->string_refs) {
+            free(context->buf);
+            free(context);
+            return NULL;
+        }
+    } else {
+        context->string_refs = NULL;
+    }
+
+    context->object_refs = create_object_context(64);
+    if (!context->object_refs) {
+        if (context->string_refs) {
+            destroy_object_context(context->string_refs);
+        }
+        free(context->buf);
+        free(context);
+        return NULL;
+    }
+
+    if (amf3) {
+        context->class_refs = create_object_context(64);
+        if (!context->class_refs) {
+            if (context->string_refs) {
+                destroy_object_context(context->string_refs);
+            }
+            if (context->object_refs) {
+                destroy_object_context(context->object_refs);
+            }
+            free(context->buf);
+            free(context);
+            return NULL;
+        }
+    } else {
+        context->class_refs = NULL;
+    }
+
+    context->use_references = 1;
     return context;
 }
 
@@ -188,9 +210,9 @@ static EncoderContext* _create_encoder_context(size_t size)
  *
  * Use this when you need to reset reference counts.
  */
-static EncoderContext* _copy_encoder_context(EncoderContext *context)
+static EncoderContext* _copy_encoder_context(EncoderContext *context, int amf3)
 {
-    EncoderContext *new_context = _create_encoder_context(context->buf_size);
+    EncoderContext *new_context = _create_encoder_context(context->buf_size, amf3);
     if (!new_context)
         return NULL;
 
@@ -229,9 +251,17 @@ static EncoderContext* _copy_encoder_context(EncoderContext *context)
 /* De-allocate an EncoderContext. */
 static int _destroy_encoder_context(EncoderContext *context)
 {
-    destroy_object_context(context->string_refs);
-    destroy_object_context(context->object_refs);
-    destroy_object_context(context->class_refs);
+    if (context->string_refs) {
+        destroy_object_context(context->string_refs);
+    }
+
+    if (context->object_refs) {
+        destroy_object_context(context->object_refs);
+    }
+
+    if (context->class_refs) {
+        destroy_object_context(context->class_refs);
+    }
 
     if (context->class_def_mapper) {
         Py_DECREF(context->class_def_mapper);
@@ -409,7 +439,7 @@ static int write_int(EncoderContext *context, PyObject *value)
         // Int is in valid AMF3 int range.
         if (_amf_write_byte(context, INT_TYPE) != 1)
             return 0;
-        return _encode_int(context, n);
+        return _encode_int((int)context, n);
     } else {
         // Int is too big, it must be encoded as a double
         if (_amf_write_byte(context, DOUBLE_TYPE) != 1)
@@ -439,9 +469,7 @@ static int serialize_unicode(EncoderContext *context, PyObject *value)
 {
     // Check for empty string
     if (PyUnicode_GET_SIZE(value) == 0) {
-        // References are never used for empty stddrings.
-        // does the index needs to be incremented ??
-        // context->string_refs->current_idx += 1;
+        // References are never used for empty strings.
         return _amf_write_byte(context, EMPTY_STRING_TYPE);
     }
 
@@ -457,19 +485,7 @@ static int serialize_unicode(EncoderContext *context, PyObject *value)
 static int encode_unicode(EncoderContext *context, PyObject *value)
 {
     PyObject *PyString_value = PyUnicode_AsUTF8String(value);
-    if (!PyString_value)
-        return 0;
-    char *char_value = PyString_AS_STRING(PyString_value);
-    int string_len = PyString_GET_SIZE(PyString_value);
-    Py_DECREF(PyString_value);
-    if (!char_value)
-        return 0;
-    
-    // Add size of string to header
-    if (!_encode_int(context, string_len << 1 | REFERENCE_BIT))
-        return 0;
-
-    return _amf_write_string_size(context, char_value, string_len);
+    return encode_string(context, PyString_value);
 }
 
 /* Serialize a PyString. */
@@ -489,16 +505,27 @@ static int serialize_string(EncoderContext *context, PyObject *value)
     return encode_string(context, value);
 }
 
-/* Encode a PyString. */
+/* Encode a PyString as UTF8. */
 static int encode_string(EncoderContext *context, PyObject *value)
 {
-    PyObject *unicode_value = PyUnicode_FromObject(value);
-    if (!unicode_value)
+    /*
+     *  TODO: The following code assumes all strings
+     *  are already UTF8 or ASCII.
+     *
+     *  Should we check to make sure,
+     *  or just let the client pick it up?
+     */
+    char *char_value = PyString_AS_STRING(value);
+    Py_ssize_t string_len = PyString_GET_SIZE(value);
+    if (!char_value)
         return 0;
 
-    int return_value = encode_unicode(context, unicode_value);
-    Py_DECREF(unicode_value);
-    return return_value;
+    // Add size of string to header
+    if (!_encode_int(context, ((int)string_len) << 1 | REFERENCE_BIT))
+        return 0;
+
+    // Write string.
+    return _amf_write_string_size(context, char_value, (int)string_len);
 }
 
 /* Encode a PyString or a PyUnicode. */
@@ -506,12 +533,11 @@ static int serialize_string_or_unicode(EncoderContext *context, PyObject *value)
 {
     if (PyUnicode_Check(value)) {
         return serialize_unicode(context, value);
-    }
-    else if (PyString_Check(value)) {
+    } else if (PyString_Check(value)) {
         return serialize_string(context, value);
     }
  
-    PyErr_SetString(amfast_EncodeError, "Attempting to encode non string/unicode as string.");
+    PyErr_SetString(amfast_EncodeError, "Attempting to serialize non string/unicode as string.");
     return 0;
 }
 
@@ -520,8 +546,7 @@ static int encode_string_or_unicode_AMF0(EncoderContext *context, PyObject *valu
 {
     if (PyUnicode_Check(value)) {
         return encode_unicode_AMF0(context, value, allow_long);
-    }
-    else if (PyString_Check(value)) {
+    } else if (PyString_Check(value)) {
         return encode_string_AMF0(context, value, allow_long);
     }
 
@@ -529,19 +554,6 @@ static int encode_string_or_unicode_AMF0(EncoderContext *context, PyObject *valu
     return 0;
 }
 
-
-/* Serialize a native C string. */
-static int _serialize_string(EncoderContext *context, char *value)
-{
-    PyObject *string_value = PyString_FromString(value);
-    if (!string_value)
-        return 0;
-
-    int return_value = serialize_string(context, string_value);
-    Py_DECREF(string_value);
-    return return_value;
-}
-
 /* Encode an ArrayCollection header. */
 static int _encode_array_collection_header(EncoderContext *context)
 {
@@ -616,7 +628,7 @@ static int encode_tuple(EncoderContext *context, PyObject *value)
 {
     // Add size of tuple to header
     Py_ssize_t value_len = PyTuple_GET_SIZE(value);
-    if (!_encode_int(context, value_len << 1 | NULL_TYPE))
+    if (!_encode_int(context, ((int)value_len) << 1 | NULL_TYPE))
         return 0;
 
     // We're never writing associative array items
@@ -624,7 +636,7 @@ static int encode_tuple(EncoderContext *context, PyObject *value)
         return 0;
 
     // Encode each value in the tuple
-    Py_ssize_t i;
+    int i;
     for (i = 0; i < value_len; i++) {
         if (!_encode(context, PyTuple_GET_ITEM(value, i)))
             return 0;
@@ -670,7 +682,7 @@ static int encode_list(EncoderContext *context, PyObject *value)
 {
     // Add size of list to header
     Py_ssize_t value_len = PyList_GET_SIZE(value);
-    if (!_encode_int(context, value_len << 1 | NULL_TYPE))
+    if (!_encode_int(context, ((int)value_len) << 1 | NULL_TYPE))
         return 0;
 
     // We're never writing associative array items
@@ -678,9 +690,15 @@ static int encode_list(EncoderContext *context, PyObject *value)
         return 0;
 
     // Encode each value in the list
-    Py_ssize_t i;
+    int i;
     for (i = 0; i < value_len; i++) {
-        if (!_encode(context, PyList_GET_ITEM(value, i)))
+        // Increment ref count in case 
+        // list is modified by someone else.
+        PyObject *list_item = PyList_GET_ITEM(value, i);
+        Py_INCREF(list_item);
+        int return_value = _encode(context, list_item);
+        Py_DECREF(list_item);
+        if (!return_value)
             return 0;
     }
 
@@ -746,18 +764,9 @@ static int serialize_dict(EncoderContext *context, PyObject *value)
 /* Encode a dict. */
 static int encode_dict(EncoderContext *context, PyObject *value)
 {
-    // Encode as anonymous object
-    if (!_amf_write_byte(context, DYNAMIC))
-        return 0;
-
-    // Anonymous object alias is an empty string
-    if (!_amf_write_byte(context, EMPTY_STRING_TYPE))
-        return 0;
-
-    // Even though this is an anonymous class,
-    // The class definition reference count needs to be incremented
-    if (!map_next_object_ref(context->class_refs, Py_None))
+    if (!serialize_class_def(context, Py_None)) {
         return 0;
+    }
     
     return _encode_dynamic_dict(context, value);
 }
@@ -806,6 +815,13 @@ static int encode_date(EncoderContext *context, PyObject *value)
         return 0;
 
     // Call python function to get datetime
+    if (!amfast_mod) {
+        amfast_mod = PyImport_ImportModule("amfast");
+        if(!amfast_mod) {
+            return 0;
+        }
+    }
+
     PyObject *epoch_func = PyObject_GetAttrString(amfast_mod, "epoch_from_date");
     if (!epoch_func)
         return 0;
@@ -871,7 +887,7 @@ static int write_reference_AMF0(EncoderContext *context, PyObject *value)
         if (!_amf_write_byte(context, REF_AMF0))
             return 0;
 
-        if (!_encode_ulong(context, idx))
+        if (!_encode_ulong(context, (unsigned int)idx))
             return 0;
 
         return 1;
@@ -894,7 +910,6 @@ static int serialize_byte_array(EncoderContext *context, PyObject *value)
 
     // Length prefix
     Py_ssize_t value_len;
-
     if (PyString_CheckExact(value)) {
         value_len = PyString_GET_SIZE(value);
     }
@@ -909,7 +924,7 @@ static int serialize_byte_array(EncoderContext *context, PyObject *value)
         return 0;
     }
 
-    if (!_encode_int(context, value_len << 1 | REFERENCE_BIT))
+    if (!_encode_int(context, ((int)value_len) << 1 | REFERENCE_BIT))
         return 0;
 
     return encode_byte_array(context, value);
@@ -1065,10 +1080,25 @@ static int encode_object(EncoderContext *context, PyObject *value)
         return 0;
     }
     
-    int static_attr_len = PyList_GET_SIZE(static_attrs);
+    Py_ssize_t static_attr_len = PySequence_Size(static_attrs);
+    if (static_attr_len == -1) {
+        Py_DECREF(class_def);
+        Py_DECREF(static_attrs);
+        return 0;
+    }
+
     int i;
     for (i = 0; i < static_attr_len; i++) {
-        if (!_encode(context, PyList_GET_ITEM(static_attrs, i))) {
+        PyObject *static_attr = PySequence_GetItem(static_attrs, i);
+        if (!static_attr) {
+            Py_DECREF(static_attrs);
+            Py_DECREF(class_def);
+            return 0;
+        }
+
+        int return_value = _encode(context, static_attr);
+        Py_DECREF(static_attr);
+        if (!return_value) {
             Py_DECREF(static_attrs);
             Py_DECREF(class_def);
             return 0;
@@ -1110,7 +1140,16 @@ static int serialize_class_def(EncoderContext *context, PyObject *value)
 /* Encode a class definition. */
 static int encode_class_def(EncoderContext *context, PyObject *value)
 {
-    int header;
+    if (value == Py_None) {
+        // Encode as anonymous object
+        if (!_amf_write_byte(context, DYNAMIC))
+            return 0;
+
+        // Anonymous object alias is an empty string
+        if (!_amf_write_byte(context, EMPTY_STRING_TYPE))
+            return 0;
+        return 1;
+    }
 
     // Encode class type
     if (!PyObject_HasAttrString(value, "CLASS_DEF")) {
@@ -1119,6 +1158,7 @@ static int encode_class_def(EncoderContext *context, PyObject *value)
     }
 
     // Determine header type
+    int header;
     if (PyObject_HasAttrString(value, "EXTERNIZEABLE_CLASS_DEF")) {
         header = EXTERNIZEABLE;
     } else if (PyObject_HasAttrString(value, "DYNAMIC_CLASS_DEF")) {
@@ -1150,19 +1190,20 @@ static int encode_class_def(EncoderContext *context, PyObject *value)
        return 0;
     }
 
-    if (!PyTuple_Check(static_attrs)) {
+    Py_ssize_t static_attr_len = PySequence_Size(static_attrs);
+    if (static_attr_len == -1) {
        Py_DECREF(class_alias);
        Py_DECREF(static_attrs);
-       PyErr_SetString(amfast_EncodeError, "ClassDef.static_attrs must be a tuple.");
        return 0;
     }
 
-    int static_attr_len = PyTuple_GET_SIZE(static_attrs);
     if (static_attr_len > (MAX_INT >> 4)) {
+       Py_DECREF(class_alias);
+       Py_DECREF(static_attrs);
        PyErr_SetString(amfast_EncodeError, "ClassDef has too many attributes.");
        return 0;
     }
-    header |= static_attr_len << 4;
+    header |= ((int)static_attr_len) << 4;
 
     if (!_encode_int(context, header)) {
        Py_DECREF(class_alias);
@@ -1180,10 +1221,15 @@ static int encode_class_def(EncoderContext *context, PyObject *value)
     // Encode static attr names
     int i;
     for (i = 0; i < static_attr_len; i++) {
-        PyObject *attr_name = PyTuple_GET_ITEM(static_attrs, i);
-        if (!attr_name)
-            return 0; 
-        if (!serialize_string_or_unicode(context, attr_name))
+        PyObject *attr_name = PySequence_GetItem(static_attrs, i);
+        if (!attr_name) {
+            Py_DECREF(static_attrs);
+            return 0;
+        }
+        
+        int return_value = serialize_string_or_unicode(context, attr_name);
+        Py_DECREF(attr_name);
+        if (!return_value)
             return 0;
     }
 
@@ -1207,7 +1253,7 @@ static int encode_bool_AMF0(EncoderContext *context, PyObject *value)
 /* Encode a PyInt in AMF0. */
 static int encode_int_AMF0(EncoderContext *context, PyObject *value)
 {
-    int long_val = PyInt_AsLong(value);
+    long long_val = PyInt_AsLong(value);
     PyObject *py_long_val = PyLong_FromLong(long_val);
     if (!py_long_val)
         return 0;
@@ -1286,7 +1332,7 @@ static int write_string_AMF0(EncoderContext *context, PyObject *value)
 /* Write a PyUnicode in AMF0. */
 static int write_unicode_AMF0(EncoderContext *context, PyObject *value)
 {
-    int string_len = PyUnicode_GET_SIZE(value);
+    Py_ssize_t string_len = PyUnicode_GET_SIZE(value);
 
     if (string_len > 65536) {
         if (!_amf_write_byte(context, LONG_STRING_AMF0))
@@ -1300,18 +1346,6 @@ static int write_unicode_AMF0(EncoderContext *context, PyObject *value)
     return encode_unicode_AMF0(context, value, 0);
 }
 
-/* Write a PyString in AMF0. */
-static int encode_string_AMF0(EncoderContext *context, PyObject *value, int allow_long)
-{
-    PyObject *unicode_value = PyUnicode_FromObject(value);
-    if (!unicode_value)
-        return 0;
-
-    int return_value = encode_unicode_AMF0(context, unicode_value, allow_long);
-    Py_DECREF(unicode_value);
-    return return_value;
-}
-
 /*
  * Encode a PyUnicode in AMF0.
  *
@@ -1325,9 +1359,24 @@ static int encode_unicode_AMF0(EncoderContext *context, PyObject *value, int all
     PyObject *PyString_value = PyUnicode_AsUTF8String(value);
     if (!PyString_value)
         return 0;
-    char *char_value = PyString_AS_STRING(PyString_value);
-    int string_len = PyString_GET_SIZE(PyString_value);
+
+    int return_value = encode_string_AMF0(context, PyString_value, allow_long);
     Py_DECREF(PyString_value);
+    return return_value;
+}
+
+/*
+ * Encode a PyString in AMF0.
+ *
+ * allow_long
+ * 1 = allow long values
+ * 0 = don't allow long values
+ * -1 = force long value
+ */
+static int encode_string_AMF0(EncoderContext *context, PyObject *value, int allow_long)
+{
+    char *char_value = PyString_AS_STRING(value);
+    Py_ssize_t string_len = PyString_GET_SIZE(value);
     if (!char_value)
         return 0;
 
@@ -1335,16 +1384,14 @@ static int encode_unicode_AMF0(EncoderContext *context, PyObject *value, int all
         PyErr_SetString(amfast_EncodeError, "Long string not allowed.");
         return 0;
     } else if (string_len > 65536 || allow_long == -1) {
-        unsigned long string_len_long = 0 + string_len;
-        if (!_encode_ulong(context, string_len_long))
+        if (!_encode_ulong(context, (unsigned int)string_len))
             return 0;
     } else {
-        unsigned short string_len_short = 0 + string_len;
-        if (!_encode_ushort(context, string_len_short))
+        if (!_encode_ushort(context, (unsigned short)string_len))
             return 0;
     }
 
-    return _amf_write_string_size(context, char_value, string_len);
+    return _amf_write_string_size(context, char_value, (size_t)string_len);
 }
 
 /* Write a PyList to AMF0. */
@@ -1359,13 +1406,19 @@ static int write_list_AMF0(EncoderContext *context, PyObject *value)
     if (!_amf_write_byte(context, ARRAY_AMF0))
         return 0;
 
-    int array_len = PyList_GET_SIZE(value);
-    if (!_encode_ulong(context, array_len))
+    Py_ssize_t array_len = PyList_GET_SIZE(value);
+    if (!_encode_ulong(context, (unsigned int)array_len))
         return 0;
 
     int i;
     for (i = 0; i < array_len; i++) {
-        if (!_encode_AMF0(context, PyList_GET_ITEM(value, i)))
+        // Increment ref count in case 
+        // list is modified by someone else.
+        PyObject *list_item = PyList_GET_ITEM(value, i);
+        Py_INCREF(list_item);
+        int return_value = _encode_AMF0(context, list_item);
+        Py_DECREF(list_item);
+        if (!return_value)
             return 0;
     }
 
@@ -1384,8 +1437,8 @@ static int write_tuple_AMF0(EncoderContext *context, PyObject *value)
     if (!_amf_write_byte(context, ARRAY_AMF0))
         return 0;
 
-    int array_len = PyTuple_GET_SIZE(value);
-    if (!_encode_ulong(context, array_len))
+    Py_ssize_t array_len = PyTuple_GET_SIZE(value);
+    if (!_encode_ulong(context, (unsigned int)array_len))
         return 0;
 
     int i;
@@ -1442,6 +1495,13 @@ static int _encode_dynamic_dict_AMF0(EncoderContext *context, PyObject *value)
 static int encode_date_AMF0(EncoderContext *context, PyObject *value)
 {
     // Call python function to get datetime
+    if (!amfast_mod) {
+        amfast_mod = PyImport_ImportModule("amfast");
+        if(!amfast_mod) {
+            return 0;
+        }
+    }
+
     PyObject *epoch_func = PyObject_GetAttrString(amfast_mod, "epoch_from_date");
     if (!epoch_func)
         return 0;
@@ -1533,6 +1593,13 @@ static int write_anonymous_object_AMF0(EncoderContext *context, PyObject *value)
 /* Get attributes from an anonymous object as a dict. */
 static PyObject* attributes_from_object(EncoderContext *context, PyObject *value)
 {
+    if (!class_def_mod) {
+        class_def_mod = PyImport_ImportModule("amfast.class_def");
+        if(!class_def_mod) {
+            return NULL;
+        }
+    }
+
     PyObject *attr_func = PyObject_GetAttrString(class_def_mod, "get_dynamic_attr_vals");
     if (!attr_func)
         return NULL;
@@ -1626,7 +1693,7 @@ static int write_object_AMF0(EncoderContext *context, PyObject *value)
             return 0;
 
         // Create new context for AMF3 encoder
-        EncoderContext *new_context = _copy_encoder_context(context);
+        EncoderContext *new_context = _copy_encoder_context(context, 1);
         int return_value = _encode(new_context, value);
         if (!return_value) {
             // AMF3 encode failed
@@ -1681,19 +1748,17 @@ static int write_object_AMF0(EncoderContext *context, PyObject *value)
         return 0;
     }
 
-    if (!PyTuple_Check(static_attr_names)) {
+    Py_ssize_t static_attr_len = PySequence_Size(static_attrs);
+    if (static_attr_len == -1) {
         Py_DECREF(class_def);
         Py_DECREF(attrs);
         Py_DECREF(static_attrs);
-        Py_DECREF(static_attr_names);
-        PyErr_SetString(amfast_EncodeError, "ClassDef.static_attrs must be a tuple.");
         return 0;
     }
-    
-    int static_attr_len = PyList_GET_SIZE(static_attrs);
+
     int i;
     for (i = 0; i < static_attr_len; i++) {
-        PyObject *static_attr_name = PyTuple_GetItem(static_attr_names, i);
+        PyObject *static_attr_name = PySequence_GetItem(static_attr_names, i);
         if (!static_attr_name) {
             Py_DECREF(class_def);
             Py_DECREF(attrs);
@@ -1701,7 +1766,7 @@ static int write_object_AMF0(EncoderContext *context, PyObject *value)
             return 0;
         }
 
-        PyObject *static_attr = PyList_GET_ITEM(static_attrs, i);
+        PyObject *static_attr = PySequence_GetItem(static_attrs, i);
         if (!static_attr) {
             Py_DECREF(class_def);
             Py_DECREF(attrs);
@@ -1721,6 +1786,7 @@ static int write_object_AMF0(EncoderContext *context, PyObject *value)
         }
     }
     Py_DECREF(static_attrs);
+    Py_DECREF(static_attr_names);
 
     // Get dynamic attrs
     if (PyObject_HasAttrString(class_def, "DYNAMIC_CLASS_DEF")) {
@@ -1804,15 +1870,19 @@ static int _encode_packet(EncoderContext *context, PyObject *value)
     if (!headers)
         return 0;
 
-    int header_count = PyList_Size(headers);
-    if (!_encode_ushort(context, header_count)) {
+    Py_ssize_t header_count = PySequence_Size(headers);
+    if (header_count == -1) {
+        Py_DECREF(headers);
+        return 0;
+    }
+    if (!_encode_ushort(context, (unsigned short)header_count)) {
         Py_DECREF(headers);
         return 0;
     }
 
     int i;
     for (i = 0; i < header_count; i++) {
-        PyObject *header = PyList_GetItem(headers, i);
+        PyObject *header = PySequence_GetItem(headers, i);
         if (!header) {
             Py_DECREF(headers);
             return 0;
@@ -1832,14 +1902,18 @@ static int _encode_packet(EncoderContext *context, PyObject *value)
     if (!messages)
         return 0;
 
-    int message_count = PyList_Size(messages);
-    if (!_encode_ushort(context, message_count)) {
+    Py_ssize_t message_count = PySequence_Size(messages);
+    if (message_count == -1) {
+        Py_DECREF(messages);
+        return 0;
+    }
+    if (!_encode_ushort(context, (unsigned short)message_count)) {
         Py_DECREF(messages);
         return 0;
     }
 
     for (i = 0; i < message_count; i++) {
-        PyObject *message = PyList_GetItem(messages, i);
+        PyObject *message = PySequence_GetItem(messages, i);
         if (!message) {
             Py_DECREF(messages);
             return 0;
@@ -1883,7 +1957,7 @@ static int encode_packet_header_AMF0(EncoderContext *context, PyObject *value)
         return 0;
 
     // Encode header value with a new context, so references are reset
-    EncoderContext *new_context = _copy_encoder_context(context);
+    EncoderContext *new_context = _copy_encoder_context(context, 0);
     return_value = _encode_AMF0(new_context, body);
     Py_DECREF(body);
     if (!return_value) {
@@ -1891,7 +1965,7 @@ static int encode_packet_header_AMF0(EncoderContext *context, PyObject *value)
         return 0;
     }
 
-    if (!_encode_ulong(context, new_context->buf_len)) {
+    if (!_encode_ulong(context, (unsigned int)new_context->buf_len)) {
         _destroy_encoder_context(new_context);
         return 0;
     }
@@ -1927,7 +2001,7 @@ static int encode_packet_message_AMF0(EncoderContext *context, PyObject *value)
         return 0;
 
     // Encode message value with a new context, so references are reset
-    EncoderContext *new_context = _copy_encoder_context(context);
+    EncoderContext *new_context = _copy_encoder_context(context, 0);
     return_value = _encode_AMF0(new_context, body);
     Py_DECREF(body);
     if (!return_value) {
@@ -1935,7 +2009,7 @@ static int encode_packet_message_AMF0(EncoderContext *context, PyObject *value)
         return 0;
     }
 
-    if (!_encode_ulong(context, new_context->buf_len)) {
+    if (!_encode_ulong(context, (unsigned int)new_context->buf_len)) {
         _destroy_encoder_context(new_context);
         return 0;
     }
@@ -2081,9 +2155,12 @@ static PyObject* encode(PyObject *self, PyObject *args, PyObject *kwargs)
         &use_legacy_xml, &amf3, &packet, &include_private, &class_def_mapper))
         return NULL;
 
-    EncoderContext *context = _create_encoder_context(1024);
-    if (!context)
+    EncoderContext *context = _create_encoder_context(1024, amf3);
+    if (!context) {
+        Py_DECREF(class_def_mapper);
+        Py_DECREF(include_private);
         return NULL;
+    }
 
     // Set defaults
     context->use_array_collections = use_array_collections;
@@ -2091,6 +2168,12 @@ static PyObject* encode(PyObject *self, PyObject *args, PyObject *kwargs)
     context->use_references = use_references;
     context->use_legacy_xml = use_legacy_xml;
 
+    if (include_private != Py_None) {
+        context->include_private = include_private;
+        Py_INCREF(context->include_private);
+        Py_DECREF(Py_False);
+    }
+
     if (class_def_mapper != Py_None) {
         // Use user supplied ClassDefMapper.
         context->class_def_mapper = class_def_mapper;
@@ -2098,27 +2181,27 @@ static PyObject* encode(PyObject *self, PyObject *args, PyObject *kwargs)
         Py_DECREF(Py_None);
     } else {
         // Create anonymous ClassDefMapper
+        Py_DECREF(Py_None);
         if (!class_def_mod) {
             class_def_mod = PyImport_ImportModule("amfast.class_def");
-            if(!class_def_mod)
+            if(!class_def_mod) {
+                _destroy_encoder_context(context);
                 return NULL;
+            }
         }
 
         PyObject *class_def = PyObject_GetAttrString(class_def_mod, "ClassDefMapper");
-        if (!class_def)
+        if (!class_def) {
+            _destroy_encoder_context(context);
             return NULL;
+        }
 
         context->class_def_mapper = PyObject_CallFunctionObjArgs(class_def, NULL);
         Py_DECREF(class_def);
-        if (!context->class_def_mapper)
+        if (!context->class_def_mapper) {
+            _destroy_encoder_context(context);
             return NULL;
-        Py_DECREF(Py_None);
-    }
-
-    if (include_private != Py_None) {
-        context->include_private = include_private;
-        Py_INCREF(context->include_private);
-        Py_DECREF(Py_False);
+        }
     }
 
     int return_value;
@@ -2135,15 +2218,8 @@ static PyObject* encode(PyObject *self, PyObject *args, PyObject *kwargs)
         return NULL;
     }
 
-    PyObject *return_obj = PyString_FromStringAndSize(context->buf, context->buf_len);
-    if (!return_obj) {
-        _destroy_encoder_context(context);
-        return NULL;
-    }
-
-    if (!_destroy_encoder_context(context))
-        return NULL;
-
+    PyObject *return_obj = PyString_FromStringAndSize(context->buf, (Py_ssize_t)context->buf_len);
+    _destroy_encoder_context(context);
     return return_obj;
 }
 
@@ -2202,14 +2278,6 @@ initencoder(void)
         return;
     }
 
-    // Include class def module
-    if (!class_def_mod) {
-        class_def_mod = PyImport_ImportModule("amfast.class_def");
-        if(!class_def_mod) {
-            return;
-        }
-    }
-
     // Determine endianness of architecture
     if (is_bigendian()) {
         big_endian = 1;
