@@ -4,6 +4,7 @@ import logging
 
 import amfast
 from amfast.class_def.sa_class_def import SaClassDef
+from amfast.class_def.code_generator import CodeGenerator
 from amfast.remoting import Service, CallableTarget
 
 import persistent
@@ -11,7 +12,7 @@ import controller
 import models
 
 def setup_gateway(gateway):
-    """configures a amfast.remoting.Gateway object."""
+    """Configures an amfast.remoting.Gateway object."""
 
     # Setup database
     schema = persistent.Schema()
@@ -23,6 +24,10 @@ def setup_gateway(gateway):
     handler.setLevel(logging.DEBUG)
     amfast.logger.addHandler(handler)
 
+    # Set Gateway options
+    gateway.use_array_collections = True
+    gateway.use_object_proxies = True
+
     # Map class aliases
     # These same aliases must be
     # registered in the client
@@ -46,3 +51,9 @@ def setup_gateway(gateway):
     service.setTarget(CallableTarget(sa_obj.insertDefaultData, 'insertDefaultData'))
     service.setTarget(CallableTarget(sa_obj.raiseException, 'raiseException'))
     gateway.service_mapper.mapService(service)
+
+    # Generate source code for mapped models
+    #coder = CodeGenerator()
+    #coder.generateFilesFromMapper(gateway.class_def_mapper, use_accessors=False,
+    #    packaged=True, constructor=False, bindable=True, extends='SAObject')
+
